@@ -18,17 +18,8 @@
 
 package gnu.trove;
 
-import java.io.IOException;
-import java.io.ObjectInputStream;
-import java.io.ObjectOutputStream;
-import java.io.Serializable;
-import java.util.Collection;
-import java.util.ConcurrentModificationException;
-import java.util.Iterator;
-import java.util.Map.Entry;
-import java.util.Map;
-import java.util.NoSuchElementException;
-import java.util.Set;
+import java.io.*;
+import java.util.*;
 
 /**
  * An implementation of the Map interface which uses an open addressed
@@ -37,12 +28,13 @@ import java.util.Set;
  * Created: Sun Nov  4 08:52:45 2001
  *
  * @author Eric D. Friedman
- * @version $Id: THashMap.java,v 1.15 2003/03/23 04:06:59 ericdf Exp $
+ * @version $Id: THashMap.java,v 1.26 2007/06/29 20:03:10 robeden Exp $
  */
-public class THashMap extends TObjectHash implements Map, Serializable {
+public class THashMap<K,V> extends TObjectHash<K> implements Map<K,V>, Externalizable {
+    static final long serialVersionUID = 1L;
 
-    /** the values of the map */
-    protected transient Object[] _values;
+    /** the values of the  map */
+    protected transient V[] _values;
 
     /**
      * Creates a new <code>THashMap</code> instance with the default
@@ -57,7 +49,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * capacity and load factor.
      * @param strategy used to compute hash codes and to compare objects.
      */
-    public THashMap(TObjectHashingStrategy strategy) {
+    public THashMap(TObjectHashingStrategy<K> strategy) {
         super(strategy);
     }
 
@@ -80,7 +72,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param initialCapacity an <code>int</code> value
      * @param strategy used to compute hash codes and to compare objects.
      */
-    public THashMap(int initialCapacity, TObjectHashingStrategy strategy) {
+    public THashMap(int initialCapacity, TObjectHashingStrategy<K> strategy) {
         super(initialCapacity, strategy);
     }
 
@@ -105,7 +97,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param loadFactor a <code>float</code> value
      * @param strategy used to compute hash codes and to compare objects.
      */
-    public THashMap(int initialCapacity, float loadFactor, TObjectHashingStrategy strategy) {
+    public THashMap(int initialCapacity, float loadFactor, TObjectHashingStrategy<K> strategy) {
         super(initialCapacity, loadFactor, strategy);
     }
 
@@ -115,7 +107,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @param map a <code>Map</code> value
      */
-    public THashMap(Map map) {
+    public THashMap(Map<K,V> map) {
         this(map.size());
         putAll(map);
     }
@@ -127,7 +119,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param map a <code>Map</code> value
      * @param strategy used to compute hash codes and to compare objects.
      */
-    public THashMap(Map map, TObjectHashingStrategy strategy) {
+    public THashMap(Map<K,V> map, TObjectHashingStrategy<K> strategy) {
         this(map.size(), strategy);
         putAll(map);
     }
@@ -135,9 +127,9 @@ public class THashMap extends TObjectHash implements Map, Serializable {
     /**
      * @return a shallow clone of this collection
      */
-    public Object clone() {
-        THashMap m = (THashMap)super.clone();
-        m._values = (Object[])this._values.clone();
+    public THashMap<K,V> clone() {
+        THashMap<K,V> m = (THashMap<K,V>)super.clone();
+        m._values = this._values.clone();
         return m;
     }
 
@@ -151,7 +143,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
         int capacity;
         
         capacity = super.setUp(initialCapacity);
-        _values = new Object[capacity];
+        _values = (V[]) new Object[capacity];
         return capacity;
     }
     
@@ -163,22 +155,21 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @return the previous value associated with <tt>key</tt>,
      * or null if none was found.
      */
-    public Object put(Object key, Object value) {
-        if (null == key) {
-            throw new NullPointerException("null keys not supported");
-        }
-        Object previous = null;
-        Object oldKey = null;
+    public V put(K key, V value) {
+        V previous = null;
+        Object oldKey;
         int index = insertionIndex(key);
+    boolean isNewMapping = true;
         if (index < 0) {
             index = -index -1;
             previous = _values[index];
+        isNewMapping = false;
         }
         oldKey = _set[index];
         _set[index] = key;
         _values[index] = value;
-        if (null == previous) {
-            postInsertHook(oldKey == null);
+        if (isNewMapping) {
+            postInsertHook(oldKey == FREE);
         }
 
         return previous;
@@ -195,11 +186,11 @@ public class THashMap extends TObjectHash implements Map, Serializable {
         if (! (other instanceof Map)) {
             return false;
         }
-        Map that = (Map)other;
+        Map<K, V> that = (Map<K, V>)other;
         if (that.size() != this.size()) {
             return false;
         }
-        return forEachEntry(new EqProcedure(that));
+        return forEachEntry(new EqProcedure<K,V>(that));
     }
 
     public int hashCode() {
@@ -207,33 +198,50 @@ public class THashMap extends TObjectHash implements Map, Serializable {
         forEachEntry(p);
         return p.getHashCode();
     }
+
+    public String toString() {
+        final StringBuffer buf = new StringBuffer("{");
+        forEachEntry(new TObjectObjectProcedure() {
+                public boolean execute(Object key, Object value) {
+                    buf.append(key);
+                    buf.append("=");
+                    buf.append(value);
+                    buf.append(", ");
+                    return true;
+                }
+            });
+        buf.append("}");
+        return buf.toString();
+    }
     
-    private final class HashProcedure implements TObjectObjectProcedure {
+    private final class HashProcedure implements TObjectObjectProcedure<K,V> {
         private int h = 0;
-        
+
         public int getHashCode() {
             return h;
         }
-        
-        public final boolean execute(Object key, Object value) {
-            h += (_hashingStrategy.computeHashCode(key) ^ (value == null ? 0 : value.hashCode()));
+
+        public final boolean execute(K key, V value) {
+            h += _hashingStrategy.computeHashCode(key) ^ (value == null ? 0 : value.hashCode());
             return true;
         }
     }
 
-    private static final class EqProcedure implements TObjectObjectProcedure {
-        private final Map _otherMap;
+    private static final class EqProcedure<K,V> implements TObjectObjectProcedure<K,V> {
+        private final Map<K,V> _otherMap;
         
-        EqProcedure(Map otherMap) {
+        EqProcedure(Map<K,V> otherMap) {
             _otherMap = otherMap;
         }
         
-        public final boolean execute(Object key, Object value) {
-            Object oValue = _otherMap.get(key);
-            if (oValue == value || (oValue != null && oValue.equals(value))) {
-                return true;
-            }
-            return false;
+        public final boolean execute(K key, V value) {
+            // Check to make sure the key is there. This avoids problems that come up with
+            // null values. Since it is only caused in that cause, only do this when the
+            // value is null (to avoid extra work).
+            if (value == null && !_otherMap.containsKey(key)) return false;
+            
+            V oValue = _otherMap.get(key);
+            return oValue == value || (oValue != null && oValue.equals(value));
         }
     }
 
@@ -244,7 +252,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @return false if the loop over the keys terminated because
      * the procedure returned false for some key.
      */
-    public boolean forEachKey(TObjectProcedure procedure) {
+    public boolean forEachKey(TObjectProcedure<K> procedure) {
         return forEach(procedure);
     }
 
@@ -255,11 +263,11 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @return false if the loop over the values terminated because
      * the procedure returned false for some value.
      */
-    public boolean forEachValue(TObjectProcedure procedure) {
-        Object[] values = _values;
+    public boolean forEachValue(TObjectProcedure<V> procedure) {
+        V[] values = _values;
         Object[] set = _set;
         for (int i = values.length; i-- > 0;) {
-            if (set[i] != null
+            if (set[i] != FREE
                 && set[i] != REMOVED
                 && ! procedure.execute(values[i])) {
                 return false;
@@ -276,13 +284,13 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @return false if the loop over the entries terminated because
      * the procedure returned false for some entry.
      */
-    public boolean forEachEntry(TObjectObjectProcedure procedure) {
+    public boolean forEachEntry(TObjectObjectProcedure<K,V> procedure) {
         Object[] keys = _set;
-        Object[] values = _values;
+        V[] values = _values;
         for (int i = keys.length; i-- > 0;) {
-            if (keys[i] != null
+            if (keys[i] != FREE
                 && keys[i] != REMOVED
-                && ! procedure.execute(keys[i],values[i])) {
+                && ! procedure.execute((K) keys[i],values[i])) {
                 return false;
             }
         }
@@ -296,18 +304,27 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param procedure determines which entries to keep
      * @return true if the map was modified.
      */
-    public boolean retainEntries(TObjectObjectProcedure procedure) {
+    public boolean retainEntries(TObjectObjectProcedure<K,V> procedure) {
         boolean modified = false;
         Object[] keys = _set;
-        Object[] values = _values;
-        for (int i = keys.length; i-- > 0;) {
-            if (keys[i] != null
-                && keys[i] != REMOVED
-                && ! procedure.execute(keys[i],values[i])) {
-                removeAt(i);
-                modified = true;
+        V[] values = _values;
+
+        // Temporarily disable compaction. This is a fix for bug #1738760
+        tempDisableAutoCompaction();
+        try {
+            for (int i = keys.length; i-- > 0;) {
+                if (keys[i] != FREE
+                    && keys[i] != REMOVED
+                    && ! procedure.execute((K) keys[i],values[i])) {
+                    removeAt(i);
+                    modified = true;
+                }
             }
         }
+        finally {
+            reenableAutoCompaction(true);
+        }
+
         return modified;
     }
 
@@ -316,11 +333,11 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @param function a <code>TObjectFunction</code> value
      */
-    public void transformValues(TObjectFunction function) {
-        Object[] values = _values;
+    public void transformValues(TObjectFunction<V,V> function) {
+        V[] values = _values;
         Object[] set = _set;
         for (int i = values.length; i-- > 0;) {
-            if (set[i] != null && set[i] != REMOVED) {
+            if (set[i] != FREE && set[i] != REMOVED) {
                 values[i] = function.execute(values[i]);
             }
         }
@@ -334,15 +351,16 @@ public class THashMap extends TObjectHash implements Map, Serializable {
     protected void rehash(int newCapacity) {
         int oldCapacity = _set.length;
         Object oldKeys[] = _set;
-        Object oldVals[] = _values;
+        V oldVals[] = _values;
 
         _set = new Object[newCapacity];
-        _values = new Object[newCapacity];
+        Arrays.fill(_set, FREE);
+        _values = (V[]) new Object[newCapacity];
 
         for (int i = oldCapacity; i-- > 0;) {
-            if(oldKeys[i] != null && oldKeys[i] != REMOVED) {
+            if(oldKeys[i] != FREE && oldKeys[i] != REMOVED) {
                 Object o = oldKeys[i];
-                int index = insertionIndex(o);
+                int index = insertionIndex((K) o);
                 if (index < 0) {
                     throwObjectContractViolation(_set[(-index -1)], o);
                 }
@@ -358,8 +376,8 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param key an <code>Object</code> value
      * @return the value of <tt>key</tt> or null if no such mapping exists.
      */
-    public Object get(Object key) {
-        int index = index(key);
+    public V get(Object key) {
+        int index = index((K) key);
         return index < 0 ? null : _values[index];
     }
 
@@ -368,12 +386,14 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      */
     public void clear() {
+        if (size() == 0) return; // optimization
+
         super.clear();
         Object[] keys = _set;
-        Object[] vals = _values;
+        V[] vals = _values;
 
         for (int i = keys.length; i-- > 0;) {
-            keys[i] = null;
+            keys[i] = FREE;
             vals[i] = null;
         }
     }
@@ -384,9 +404,9 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param key an <code>Object</code> value
      * @return an <code>Object</code> value
      */
-    public Object remove(Object key) {
-        Object prev = null;
-        int index = index(key);
+    public V remove(Object key) {
+        V prev = null;
+        int index = index((K) key);
         if (index >= 0) {
             prev = _values[index];
             removeAt(index);    // clear key,state; adjust size
@@ -400,8 +420,8 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * @param index an <code>int</code> value
      */
     protected void removeAt(int index) {
-        super.removeAt(index);  // clear key, state; adjust size
         _values[index] = null;
+        super.removeAt(index);  // clear key, state; adjust size
     }
 
     /**
@@ -409,7 +429,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @return a <code>Collection</code> value
      */
-    public Collection values() {
+    public Collection<V> values() {
         return new ValueView();
     }
 
@@ -418,7 +438,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @return a <code>Set</code> value
      */
-    public Set keySet() {
+    public Set<K> keySet() {
         return new KeyView();
     }
 
@@ -427,7 +447,7 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @return a <code>Set</code> value
      */
-    public Set entrySet() {
+    public Set<Map.Entry<K,V>> entrySet() {
         return new EntryView();
     }
 
@@ -439,20 +459,20 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      */
     public boolean containsValue(Object val) {
         Object[] set = _set;
-        Object[] vals = _values;
+        V[] vals = _values;
 
         // special case null values so that we don't have to
         // perform null checks before every call to equals()
         if (null == val) {
             for (int i = vals.length; i-- > 0;) {
-                if ((set[i] != null && set[i] != REMOVED) &&
+                if ((set[i] != FREE && set[i] != REMOVED) &&
                     val == vals[i]) {
                     return true;
                 }
             }
         } else {
             for (int i = vals.length; i-- > 0;) {
-                if ((set[i] != null && set[i] != REMOVED) &&
+                if ((set[i] != FREE && set[i] != REMOVED) &&
                     (val == vals[i] || val.equals(vals[i]))) {
                     return true;
                 }
@@ -476,12 +496,12 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      *
      * @param map a <code>Map</code> value
      */
-    public void putAll(Map map) {
+    public void putAll(Map<? extends K, ? extends V> map) {
         ensureCapacity(map.size());
         // could optimize this for cases when map instanceof THashMap
-        for (Iterator i = map.entrySet().iterator(); i.hasNext();) {
-            Map.Entry e = (Map.Entry)i.next();
-            put(e.getKey(),e.getValue());
+      for (Iterator<? extends Map.Entry<? extends K,? extends V>> i = map.entrySet().iterator(); i.hasNext();) {
+        Map.Entry<? extends K,? extends V> e = i.next();
+        put(e.getKey(),e.getValue());
         }
     }
 
@@ -489,33 +509,34 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * a view onto the values of the map.
      *
      */
-    protected class ValueView extends KeyView {
-        public Iterator iterator() {
-            return new THashIterator(THashMap.this) {
-                    protected Object objectAtIndex(int index) {
-                        return _values[index];
-                    }
-                };
+    protected class ValueView extends MapBackedView<V> {
+        public Iterator<V> iterator() {
+            return new THashIterator<V>(THashMap.this) {
+                protected V objectAtIndex(int index) {
+                    return _values[index];
+                }
+            };
         }
 
-        public boolean contains(Object value) {
+        public boolean containsElement(V value) {
             return containsValue(value);
         }
 
-        public boolean remove(Object value) {
-            boolean changed = false;
+        public boolean removeElement(V value) {
             Object[] values = _values;
             Object[] set = _set;
             
             for (int i = values.length; i-- > 0;) {
-                if ((set[i] != null && set[i] != REMOVED) &&
+                if ((set[i] != FREE && set[i] != REMOVED) &&
                     value == values[i] ||
                     (null != values[i] && values[i].equals(value))) {
+
                     removeAt(i);
-                    changed = true;
+                    return true;
                 }
             }
-            return changed;
+
+            return false;
         }
     }
 
@@ -523,12 +544,22 @@ public class THashMap extends TObjectHash implements Map, Serializable {
      * a view onto the entries of the map.
      *
      */
-    protected class EntryView extends KeyView {
-        public Iterator iterator() {
+    protected class EntryView extends MapBackedView<Map.Entry<K,V>> {
+        private final class EntryIterator extends THashIterator<Map.Entry<K,V>> {
+            EntryIterator(THashMap<K,V> map) {
+                super(map);
+            }
+
+            public Entry objectAtIndex(final int index) {
+                return new Entry((K) _set[index], _values[index], index);
+            }
+        }
+
+        public Iterator<Map.Entry<K,V>> iterator() {
             return new EntryIterator(THashMap.this);
         }
 
-        public boolean remove(Object entry) {
+        public boolean removeElement(Map.Entry<K,V> entry) {
             // have to effectively reimplement Map.remove here
             // because we need to return true/false depending on
             // whether the removal took place.  Since the Entry's
@@ -538,10 +569,10 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             //
             // Note also that the deletion is only legal if
             // both the key and the value match.
-            Object key, val;
+            Object val;
             int index;
 
-            key = keyForEntry(entry);
+            K key = keyForEntry(entry);
             index = index(key);
             if (index >= 0) {
                 val = valueForEntry(entry);
@@ -554,40 +585,40 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             return false;
         }
 
-        public boolean contains(Object entry) {
+        public boolean containsElement(Map.Entry<K,V> entry) {
             Object val = get(keyForEntry(entry));
-            Object entryValue = ((Map.Entry)entry).getValue();
+            Object entryValue = entry.getValue();
             return entryValue == val ||
                 (null != val && val.equals(entryValue));
         }
 
-        protected Object valueForEntry(Object entry) {
-            return ((Map.Entry)entry).getValue();
+        protected V valueForEntry(Map.Entry<K,V> entry) {
+            return entry.getValue();
         }
 
-        protected Object keyForEntry(Object entry) {
-            return ((Map.Entry)entry).getKey();
+        protected K keyForEntry(Map.Entry<K,V> entry) {
+            return entry.getKey();
         }
     }
-    
-    /**
-     * a view onto the keys of the map.
-     */
-    protected class KeyView implements Set {
-        public Iterator iterator() {
-            return new TObjectHashIterator(THashMap.this);
-        }
 
-        public boolean remove(Object key) {
-            return (null != THashMap.this.remove(key));
-        }
+    private abstract class MapBackedView<E> extends AbstractSet<E>
+        implements Set<E>, Iterable<E> {
+        
+        public abstract Iterator<E> iterator();
+
+        public abstract boolean removeElement(E key);
+
+        public abstract boolean containsElement(E key);
 
         public boolean contains(Object key) {
-            return THashMap.this.contains(key);
+            return containsElement((E) key);
         }
 
+        public boolean remove(Object o) {
+            return removeElement((E) o);
+        }
 
-        public boolean containsAll(Collection collection) {
+        public boolean containsAll(Collection<?> collection) {
             for (Iterator i = collection.iterator(); i.hasNext();) {
                 if (! contains(i.next())) {
                     return false;
@@ -596,21 +627,11 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             return true;
         }
 
-        public boolean removeAll(Collection collection) {
-            boolean changed = false;
-            for (Iterator i = collection.iterator(); i.hasNext();) {
-                if (remove(i.next())) {
-                    changed = true;
-                }
-            }
-            return changed;
-        }
-
         public void clear() {
             THashMap.this.clear();
         }
 
-        public boolean add(Object obj) {
+        public boolean add(E obj) {
             throw new UnsupportedOperationException();
         }
 
@@ -626,14 +647,15 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             return result;
         }
 
-        public Object[] toArray(Object[] a) {
+          public <T> T[] toArray(T[] a) {
             int size = size();
             if (a.length < size)
-                a = (Object[])java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
+                a = (T[]) java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
 
-            Iterator it = iterator();
+            Iterator<E> it = iterator();
+            Object[] result = a;
             for (int i=0; i<size; i++) {
-                a[i] = it.next();
+                result[i] = it.next();
             }
 
             if (a.length > size) {
@@ -647,11 +669,11 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             return THashMap.this.isEmpty();
         }
 
-        public boolean addAll(Collection collection) {
+        public boolean addAll(Collection<? extends E> collection) {
             throw new UnsupportedOperationException();
         }
 
-        public boolean retainAll(Collection collection) {
+        public boolean retainAll(Collection<?> collection) {
             boolean changed = false;
             Iterator i = iterator();
             while (i.hasNext()) {
@@ -662,38 +684,53 @@ public class THashMap extends TObjectHash implements Map, Serializable {
             }
             return changed;
         }
+    }
 
-        protected final class EntryIterator extends THashIterator {
-            EntryIterator(THashMap map) {
-                super(map);
-            }
-            
-            public Object objectAtIndex(final int index) {
-                return new Entry(_set[index], _values[index], index);
-            }
+    /**
+     * a view onto the keys of the map.
+     */
+    protected class KeyView extends MapBackedView<K> {
+        public Iterator<K> iterator() {
+            return new TObjectHashIterator<K>(THashMap.this);
+        }
+
+        public boolean removeElement(K key) {
+            return null != THashMap.this.remove(key);
+        }
+
+        public boolean containsElement(K key) {
+            return THashMap.this.contains(key);
         }
     }
 
-    final class Entry implements Map.Entry {
-        private final Object key;
-        private Object val;
+    final class Entry implements Map.Entry<K,V> {
+        private K key;
+        private V val;
         private final int index;
 
-        Entry(final Object key, Object value, final int index) {
+        Entry(final K key, V value, final int index) {
             this.key = key;
             this.val = value;
             this.index = index;
         }
         
-        public Object getKey() {
+        void setKey(K aKey) {
+            this.key = aKey;
+        }
+
+        void setValue0(V aValue) {
+            this.val = aValue;
+        }
+        
+        public K getKey() {
             return key;
         }
         
-        public Object getValue() {
+        public V getValue() {
             return val;
         }
         
-        public Object setValue(Object o) {
+        public V setValue(V o) {
             if (_values[index] != val) {
                 throw new ConcurrentModificationException();
             }
@@ -703,30 +740,51 @@ public class THashMap extends TObjectHash implements Map, Serializable {
                                 // setValue is called again
             return o;
         }
+
+        public boolean equals(Object o) {
+            if (o instanceof Map.Entry) {
+                Map.Entry e1 = this;
+                Map.Entry e2 = (Map.Entry) o;
+                return (e1.getKey()==null ? e2.getKey()==null : e1.getKey().equals(e2.getKey()))
+                    && (e1.getValue()==null ? e2.getValue()==null : e1.getValue().equals(e2.getValue()));
+            }
+            return false;
+        }
+
+        public int hashCode() {
+            return (getKey()==null ? 0 : getKey().hashCode()) ^ (getValue()==null ? 0 : getValue().hashCode());
+        }
     }
 
-    private void writeObject(ObjectOutputStream stream)
-        throws IOException {
-        stream.defaultWriteObject();
 
-        // number of entries
-        stream.writeInt(_size);
+    public void writeExternal( ObjectOutput out ) throws IOException {
+        // VERSION
+        out.writeByte( 0 );
 
-        SerializationProcedure writeProcedure = new SerializationProcedure(stream);
+        // NUMBER OF ENTRIES
+        out.writeInt( _size );
+
+        // ENTRIES
+        SerializationProcedure writeProcedure = new SerializationProcedure( out );
         if (! forEachEntry(writeProcedure)) {
             throw writeProcedure.exception;
         }
     }
 
-    private void readObject(ObjectInputStream stream)
+    public void readExternal( ObjectInput in )
         throws IOException, ClassNotFoundException {
-        stream.defaultReadObject();
 
-        int size = stream.readInt();
-        setUp(size);
+        // VERSION
+        in.readByte();
+
+        // NUMBER OF ENTRIES
+        int size = in.readInt();
+        setUp( size );
+
+        // ENTRIES
         while (size-- > 0) {
-            Object key = stream.readObject();
-            Object val = stream.readObject();
+            K key = (K) in.readObject();
+            V val = (V) in.readObject();
             put(key, val);
         }
     }
