@@ -19,9 +19,8 @@
 
 package gnu.trove;
 
-import java.io.Serializable;
+import java.io.*;
 import java.util.AbstractSequentialList;
-import java.util.Iterator;
 import java.util.ListIterator;
 import java.util.NoSuchElementException;
 
@@ -52,17 +51,20 @@ import java.util.NoSuchElementException;
  * </p>
  *
  * @author Eric D. Friedman
- * @version $Id: TLinkedList.java,v 1.4 2002/04/08 02:02:28 ericdf Exp $
+ * @version $Id: TLinkedList.java,v 1.10 2007/06/29 21:42:20 robeden Exp $
  * @see gnu.trove.TLinkable
  */
 
-public class TLinkedList extends AbstractSequentialList
-    implements Serializable {
+public class TLinkedList<T extends TLinkable> extends AbstractSequentialList<T>
+    implements Externalizable {
+
+    static final long serialVersionUID = 1L;
+
 
     /** the head of the list */
-    protected TLinkable _head;
+    protected T _head;
     /** the tail of the list */
-    protected TLinkable _tail;
+    protected T _tail;
     /** the number of elements in the list */
     protected int _size = 0;
     
@@ -86,7 +88,7 @@ public class TLinkedList extends AbstractSequentialList
      * @param index an <code>int</code> value
      * @return a <code>ListIterator</code> value
      */
-    public ListIterator listIterator(int index) {
+    public ListIterator<T> listIterator(int index) {
         return new IteratorImpl(index);
     }
 
@@ -107,7 +109,7 @@ public class TLinkedList extends AbstractSequentialList
      * @param index an <code>int</code> value
      * @param linkable an object of type TLinkable
      */
-    public void add(int index, Object linkable) {
+    public void add(int index, T linkable) {
         if (index < 0 || index > size()) {
             throw new IndexOutOfBoundsException("index:" + index);
         }
@@ -120,7 +122,7 @@ public class TLinkedList extends AbstractSequentialList
      * @param linkable an object of type TLinkable
      * @return always true
      */
-    public boolean add(Object linkable) {
+    public boolean add(T linkable) {
         insert(_size, linkable);
         return true;
     }
@@ -130,7 +132,7 @@ public class TLinkedList extends AbstractSequentialList
      *
      * @param linkable an object of type TLinkable
      */
-    public void addFirst(Object linkable) {
+    public void addFirst(T linkable) {
         insert(0, linkable);
     }
 
@@ -139,7 +141,7 @@ public class TLinkedList extends AbstractSequentialList
      *
      * @param linkable an object of type TLinkable
      */
-    public void addLast(Object linkable) {
+    public void addLast(T linkable) {
         insert(size(), linkable);
     }
 
@@ -194,10 +196,10 @@ public class TLinkedList extends AbstractSequentialList
     public Object[] toUnlinkedArray() {
         Object[] o = new Object[_size];
         int i = 0;
-        for (TLinkable link = _head, tmp = null; link != null; i++) {
+        for (T link = _head, tmp = null; link != null; i++) {
             o[i] = link;
             tmp = link;
-            link = link.getNext();
+            link = (T) link.getNext();
             tmp.setNext(null); // clear the links
             tmp.setPrevious(null);
         }
@@ -221,12 +223,53 @@ public class TLinkedList extends AbstractSequentialList
         return false;
     }
 
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public T get(int index) {
+        // Short circuit for easy cases
+        if (index == 0) return _head;
+        if (index == _size - 1) return _tail;
+
+        // Blow out for bogus values
+        if (index < 0 || index >= _size ) {
+            throw new IndexOutOfBoundsException("Index: " + index + ", Size: " + _size);
+        }
+
+        // Determine if it's better to get there from the front or the back
+        if (index > (_size >> 1)) {
+            int position = _size - 1;
+            T node = _tail;
+
+            while ( position > index ) {
+                node = ( T ) node.getPrevious();
+                position--;
+            }
+
+            return node;
+        }
+        else {
+            int position = 0;
+            T node = _head;
+
+            while( position < index ) {
+                node = ( T ) node.getNext();
+                position++;
+            }
+
+            return node;
+        }
+    }
+
+
     /**
      * Returns the head of the list
      *
      * @return an <code>Object</code> value
      */
-    public Object getFirst() {
+    public T getFirst() {
         return _head;
     }
 
@@ -235,18 +278,52 @@ public class TLinkedList extends AbstractSequentialList
      *
      * @return an <code>Object</code> value
      */
-    public Object getLast() {
+    public T getLast() {
         return _tail;
     }
 
+
+    /**
+     * Return the node following the given node. This method exists for two reasons:
+     * <ol>
+     *   <li>It's really not recommended that the methods implemented by TLinkable be
+     *       called directly since they're used internally by this class.</li>
+     *   <li>This solves problems arising from generics when working with the linked
+     *       objects directly.</li>
+     * </ol>
+     * <p>
+     * NOTE: this should only be used with nodes contained in the list. The results are
+     * undefined with anything else.
+     */
+    public T getNext( T current ) {
+        return ( T ) current.getNext();
+    }
+
+    /**
+     * Return the node preceding the given node. This method exists for two reasons:
+     * <ol>
+     *   <li>It's really not recommended that the methods implemented by TLinkable be
+     *       called directly since they're used internally by this class.</li>
+     *   <li>This solves problems arising from generics when working with the linked
+     *       objects directly.</li>
+     * </ol>
+     * <p>
+     * NOTE: this should only be used with nodes contained in the list. The results are
+     * undefined with anything else.
+     */
+    public T getPrevious( T current ) {
+        return ( T ) current.getPrevious();
+    }
+
+
     /**
      * Remove and return the first element in the list.
      *
      * @return an <code>Object</code> value
      */
-    public Object removeFirst() {
-        TLinkable o = _head;
-        TLinkable n = o.getNext();
+    public T removeFirst() {
+        T o = _head;
+        T n = (T) o.getNext();
         o.setNext(null);
 
         if (null != n) {
@@ -265,9 +342,9 @@ public class TLinkedList extends AbstractSequentialList
      *
      * @return an <code>Object</code> value
      */
-    public Object removeLast() {
-        TLinkable o = _tail;
-        TLinkable prev = o.getPrevious();
+    public T removeLast() {
+        T o = _tail;
+        T prev = (T) o.getPrevious();
         o.setPrevious(null);
 
         if (null != prev) {
@@ -286,8 +363,8 @@ public class TLinkedList extends AbstractSequentialList
      * @param index an <code>int</code> value
      * @param linkable an object of type TLinkable
      */
-    protected void insert(int index, Object linkable) {
-        TLinkable newLink = (TLinkable)linkable;
+    protected void insert(int index, T linkable) {
+        T newLink = linkable;
 
         if (_size == 0) {
             _head = _tail = newLink; // first insertion
@@ -300,31 +377,14 @@ public class TLinkedList extends AbstractSequentialList
             newLink.setPrevious(_tail);
             _tail = newLink;
         } else {
-            TLinkable prior = null, post = null;
+            T node = get(index);
 
-            // looking at the size of the list, we decide whether
-            // it's faster to reach `index' by traversing the
-            // list from the front or the back.
-            if (index > (_size >> 1)) { // insert in 2nd half
-                // work from the tail
-                int pos = _size -1;
-                for (prior = _tail; pos > index; pos--) {
-                    prior = prior.getPrevious();
-                }
-            } else {                // insert in 1st half
-                // work from the head
-                int pos = 0;
-                for (prior = _head; pos < index; pos++) {
-                    prior = prior.getNext();
-                }
-            }
-            post = prior.getNext();
-            // insert newlink
-            newLink.setNext(post);
-            newLink.setPrevious(prior);
-            // adjust adjacent pointers
-            post.setPrevious(newLink);
-            prior.setNext(newLink);
+            T before = (T) node.getPrevious();
+            if (before != null) before.setNext(linkable);
+
+            linkable.setPrevious(before);
+            linkable.setNext(node);
+            node.setPrevious(linkable);
         }
         _size++;
     }
@@ -340,11 +400,11 @@ public class TLinkedList extends AbstractSequentialList
      */
     public boolean remove(Object o) {
         if (o instanceof TLinkable) {
-            TLinkable p, n;
+            T p, n;
             TLinkable link = (TLinkable)o;
 
-            p = link.getPrevious();
-            n = link.getNext();
+            p = (T) link.getPrevious();
+            n = (T) link.getNext();
 
             if (n == null && p == null) { // emptying the list
                 _head = _tail = null;
@@ -381,7 +441,7 @@ public class TLinkedList extends AbstractSequentialList
      * @param newElement a <code>TLinkable</code> value to be added to
      * the list.
      */
-    public void addBefore(TLinkable current, TLinkable newElement) {
+    public void addBefore(T current, T newElement) {
         if (current == _head) {
             addFirst(newElement);
         } else if (current == null) {
@@ -396,14 +456,91 @@ public class TLinkedList extends AbstractSequentialList
         }
     }
 
+    /**
+     * Inserts newElement into the list immediately after current.
+     * All elements to the left of and including current are shifted
+     * over.
+     *
+     * @param current a <code>TLinkable</code> value currently in the list.
+     * @param newElement a <code>TLinkable</code> value to be added to
+     * the list.
+     */
+    public void addAfter(T current, T newElement) {
+        if (current == _tail) {
+            addLast(newElement);
+        } else if (current == null) {
+            addFirst(newElement);
+        } else {
+            TLinkable n = current.getNext();
+            newElement.setPrevious(current);
+            newElement.setNext(n);
+            current.setNext(newElement);
+            n.setPrevious(current);
+            _size++;
+        }
+    }
+
+
+    /**
+     * Executes <tt>procedure</tt> for each entry in the list.
+     *
+     * @param procedure a <code>TObjectProcedure</code> value
+     * @return false if the loop over the values terminated because
+     * the procedure returned false for some value.
+     */
+    public boolean forEachValue(TObjectProcedure<T> procedure) {
+        T node = _head;
+        while ( node != null ) {
+            boolean keep_going = procedure.execute( node );
+            if ( !keep_going ) return false;
+
+            node = ( T ) node.getNext();
+        }
+
+        return true;
+    }
+
+
+
+    public void writeExternal( ObjectOutput out ) throws IOException {
+        // VERSION
+        out.writeByte(0);
+
+        // NUMBER OF ENTRIES
+        out.writeInt(_size);
+
+        // HEAD
+        out.writeObject(_head);
+
+        // TAIL
+        out.writeObject(_head);
+    }
+
+    public void readExternal( ObjectInput in )
+        throws IOException, ClassNotFoundException {
+
+        // VERSION
+        in.readByte();
+
+        // NUMBER OF ENTRIED
+        _size = in.readInt();
+
+        // HEAD
+        _head = (T) in.readObject();
+
+        // TAIL
+        _tail = (T) in.readObject();
+    }
+
+
     /**
      * A ListIterator that supports additions and deletions.
      *
      */
-    protected final class IteratorImpl implements ListIterator {
+    protected final class IteratorImpl implements ListIterator<T> {
         private int _nextIndex = 0;
-        private TLinkable _next;
-        private TLinkable _lastReturned;
+        private T _next;
+        private T _lastReturned;
 
         /**
          * Creates a new <code>Iterator</code> instance positioned at
@@ -415,7 +552,7 @@ public class TLinkedList extends AbstractSequentialList
             if (position < 0 || position > _size) {
                 throw new IndexOutOfBoundsException();
             }
-            
+
             _nextIndex = position;
             if (position == 0) {
                 _next = _head;
@@ -424,12 +561,12 @@ public class TLinkedList extends AbstractSequentialList
             } else if (position < (_size >> 1)) {
                 int pos = 0;
                 for (_next = _head; pos < position; pos++) {
-                    _next = _next.getNext();
+                    _next = (T) _next.getNext();
                 }
             } else {
                 int pos = _size - 1;
                 for (_next = _tail; pos > position; pos--) {
-                    _next = _next.getPrevious();
+                    _next = (T) _next.getPrevious();
                 }
             }
         }
@@ -440,14 +577,14 @@ public class TLinkedList extends AbstractSequentialList
          *
          * @param linkable an object of type TLinkable
          */
-        public final void add(Object linkable) {
+        public final void add(T linkable) {
             _lastReturned = null;
             _nextIndex++;
 
             if (_size == 0) {
                 TLinkedList.this.add(linkable);
             } else {
-                TLinkedList.this.addBefore(_next, (TLinkable)linkable);
+                TLinkedList.this.addBefore(_next, linkable);
             }
         }
 
@@ -476,13 +613,13 @@ public class TLinkedList extends AbstractSequentialList
          * @return an <code>Object</code> value
          * @exception NoSuchElementException if there is no next element
          */
-        public final Object next() {
+        public final T next() {
             if (_nextIndex == _size) {
                 throw new NoSuchElementException();
             }
 
             _lastReturned = _next;
-            _next = _next.getNext();
+            _next = (T) _next.getNext();
             _nextIndex++;
             return _lastReturned;
         }
@@ -504,7 +641,7 @@ public class TLinkedList extends AbstractSequentialList
          * @return an <code>Object</code> value
          * @exception NoSuchElementException if there is no previous element.
          */
-        public final Object previous() {
+        public final T previous() {
             if (_nextIndex == 0) {
                 throw new NoSuchElementException();
             }
@@ -512,9 +649,9 @@ public class TLinkedList extends AbstractSequentialList
             if (_nextIndex == _size) {
                 _lastReturned = _next = _tail;
             } else {
-                _lastReturned = _next = _next.getPrevious();
+                _lastReturned = _next = (T) _next.getPrevious();
             }
-            
+
             _nextIndex--;
             return _lastReturned;
         }
@@ -544,7 +681,7 @@ public class TLinkedList extends AbstractSequentialList
             if (_lastReturned != _next) {
                 _nextIndex--;
             }
-            _next = _lastReturned.getNext();
+            _next = (T) _lastReturned.getNext();
             TLinkedList.this.remove(_lastReturned);
             _lastReturned = null;
         }
@@ -555,11 +692,11 @@ public class TLinkedList extends AbstractSequentialList
          *
          * @param linkable an object of type TLinkable
          */
-        public final void set(Object linkable) {
+        public final void set(T linkable) {
             if (_lastReturned == null) {
                 throw new IllegalStateException();
             }
-            TLinkable l = (TLinkable)linkable;
+            T l = linkable;
 
             // need to check both, since this could be the only
             // element in the list.
@@ -581,9 +718,9 @@ public class TLinkedList extends AbstractSequentialList
          * @param from a <code>TLinkable</code> value
          * @param to a <code>TLinkable</code> value
          */
-        private void swap(TLinkable from, TLinkable to) {
-            TLinkable p = from.getPrevious();
-            TLinkable n = from.getNext();
+        private void swap(T from, T to) {
+            T p = (T) from.getPrevious();
+            T n = (T) from.getNext();
 
             if (null != p) {
                 to.setPrevious(p);
